"""A simple emulated reference device that includes a blank non-controller tile."""

from ..reference import ReferenceDevice


class DemoReferenceDevice(ReferenceDevice):
    """A basic reference device with a reference controller."""

    pass
